# Project: SILECS
# Author: S.Magnoni/ BE-CO
# Date: August 2012
# Last modification: Jannuary 2015
#
# Description:
# This module implement the methods to build each components of the sources.
# A component source consists of one 'parameterizable' text that can
# be specialized on call to append the string to be stored in the file.
#

import string
import time


#-------------------------------------------------------------------------
# Hash-Table

whichDIGIFormat = {
    'uint8'   : 'uint16_t',
    'int8'    : 'int16_t',
    'uint16'  : 'uint16_t',
    'int16'   : 'int16_t',
    'uint32'  : 'uint32_t',
    'int32'   : 'int32_t',
    'float32' : 'float',
    'uint64'  : '8',
    'int64'   : '8',
    'string'  : 'uint8_t',    
#    'float64' : '8',    not supported by microcontroller
    'date'    : 'dt',
       'char'    :'int16_t',
    'byte'    :'uint16_t',
    'word'    :'uint16_t',
    'dword'    :'uint32_t',
    'int'    :'int16_t',
    'dint'    :'int32_t',
    'real'    :'float',
    'dt'    :'dt'
}

#=========================================================================
# Utils
#=========================================================================

def toWordArray(strg):
    wordArray = ''
    for i in range(0, len(strg), 2):
        lsb = ord(strg[i])
        try: hsb = ord(strg[i+1]) 
        except: hsb = 0
        wordArray += "0x%02x%02x," %(hsb,lsb)
    return wordArray

#=========================================================================
# C Source template (.c file)
#=========================================================================
header = string.Template("""
/* +-------------------------------------------------------------------
 * | Copyright CERN 2015
 * | SILECS - BE/CO-SRC
 * | April 2015
 * +-------------------------------------------------------------------
 *
 * Release : ${silecsRelease}
 *
 * The following code has been automatically generated by SILECS.
 *
 * N.B: This file relies on the existence of explicit C data type such
 * as int8_t, uint8_t, int16_t, etc....
 * If your compiler does not support them natively please implement 
 * them by including the data type definition provided on the SILECS
 * web page before including this header file.
 */

#define MODBUS_START_ADDRESS ${baseAddress}

/*---------------------------------------------------------------------
 * DT 
 * data type definition and related utilities
 *---------------------------------------------------------------------
 */

typedef struct
{
    uint8_t sc_100;     // second cent
    uint8_t sc;            // second
    uint8_t mn;         // minute
    uint8_t hh;         // hour
    uint8_t dd;         // day
    uint8_t mm;         // month
    uint8_t yy1;        // year
    uint8_t yy2;        // year2
} dt;

#define _frombcd(a) (int)(((a>>4)*10)+(a&0x0F))
#define _tobcd(a)   (((unsigned char)((a)/10)<<4)+((a)%10))

void SILECS_set_dt(int8_t sc_100 ,int8_t sc, int8_t mn,int8_t hh,int8_t dd,int8_t mm,int32_t yy, dt *date)
{
    date->sc_100 = sc_100;
    date->sc     = _tobcd(sc);
    date->mn     = _tobcd(mn);
    date->hh     = _tobcd(hh);
    date->dd     = _tobcd(dd);
    date->mm     = _tobcd(mm);
    date->yy2    = _tobcd((int8_t)(yy/100));
    date->yy1    = _tobcd((int8_t)(yy%100));
}

""")

#=========================================================================
# DATA TYPE DEFINITION
#=========================================================================
firstBlockUDT = string.Template("""
/*---------------------------------------------------------------------
 * ${className} / v${classVersion}
 * BLOCK Type definition
 *---------------------------------------------------------------------
 */
""")

blockUDT = string.Template("""
typedef struct 
{
$regList
} _${className}_${blockName};
""")

regScalar = string.Template("""    ${regFormat} ${regName};
""")

regArray  = string.Template("""    ${regFormat} ${regName}[${strLen}];
""")

regArray2d  = string.Template("""    ${regFormat} ${regName}[${regDim}][${strLen}];
""")

stringArray = string.Template("""    ${regFormat} ${regName}[${regDim}][${regDim2}][${strLen}];
""")

#=========================================================================
# INSTANTIATION
#=========================================================================

allocationComment = string.Template("""
/*---------------------------------------------------------------------
 * MEMORY ALLOCATION
 * PROTOCOL: ${plcProtocol}
 *---------------------------------------------------------------------
 */
""")

#======= DEVICE MODE =================
deviceModeBlockInstantiation = string.Template("""        _${className}_${blockName} ${blockName};
""")

deviceModeClass_deviceNumber = string.Template("""
    struct {
${blockList}
    } ${className}_device[NB_${className}_DEVICE];
""")
                                               
deviceModeClass_deviceList = string.Template("""
    struct {
${blockList}
    } ${deviceList};
""")

deviceMode_dataInstantiation = string.Template("""
typedef struct {
    ${classList}
} _SILECS_DATA_SEGMENT;

#define SILECS_DATA_SEGMENT_MODBUS_SIZE (sizeof(_SILECS_DATA_SEGMENT)/2)

union silecsData {
    _SILECS_DATA_SEGMENT data;
    uint16_t array[SILECS_DATA_SEGMENT_MODBUS_SIZE];
} silecsData;

""")

#======= BLOCK MODE =================
blockModeDeviceInstantiation_deviceList = string.Template("""        _${className}_${blockName} ${deviceName};
""")

blockModeBlockInstantiation = string.Template("""
    struct {
${deviceList}    } ${className}_${blockName};
""")

blockMode_dataInstantiation = string.Template("""
typedef struct {
    ${classList}
} _SILECS_DATA_SEGMENT;

#define SILECS_DATA_SEGMENT_MODBUS_SIZE (sizeof(_SILECS_DATA_SEGMENT)/2)

union modbus_data {
    _SILECS_DATA_SEGMENT data;
    uint16_t array[SILECS_DATA_SEGMENT_MODBUS_SIZE];
} silecsData;

""")

#=========================================================================
# init
#=========================================================================

globalAllocation = """
    SILECS_INSTANCE_DATA data;
    uint16_t modbus_data = & data;
"""

initFunctionDeviceMode = string.Template("""
/* Initialization function */
int SILECS_init()
{
    /* Silecs version initialization */
    strcpy((unsigned char *)silecsData.data.SilecsHeader_device[0].hdrBlk._version, "${version}");
    
    /* Silecs checksum initialization */
    silecsData.data.SilecsHeader_device[0].hdrBlk._checksum  = ${checksum};
    
    /* Silecs user initialization */
    strcpy((unsigned char *)silecsData.data.SilecsHeader_device[0].hdrBlk._user, "${user}");
    
    /* Silecs date initialization */
    SILECS_set_dt(${dt6},${dt5},${dt4},${dt3},${dt2},${dt1},${dt0},&silecsData.data.SilecsHeader_device[0].hdrBlk._date);
}

""")

initFunctionBlockMode = string.Template("""
/* Initialization function */
int SILECS_init()
{
    /* Silecs version initialization */
    strcpy((unsigned char *)silecsData.data.SilecsHeader_hdrBlk.device[0]._version, "${version}");
    
    /* Silecs checksum initialization */
    silecsData.data.SilecsHeader_hdrBlk.device[0]._checksum  = ${checksum};
    
    /* Silecs user initialization */
    strcpy((unsigned char *)silecsData.data.SilecsHeader_hdrBlk.device[0]._user, "${user}");
    
    /* Silecs date initialization */
    SILECS_set_dt(${dt6},${dt5},${dt4},${dt3},${dt2},${dt1},${dt0},&silecsData.data.SilecsHeader_hdrBlk.device[0]._date);
}

""")

#=========================================================================
# Example generation
#=========================================================================

instantiationExample = """/*
 * Automatically generated Addressing example
 *"""

deviceModeDeviceListExample = string.Template("""
 * This example shows how to address the register ${registerName} of block ${blockName}
 * of device ${deviceLabel} of the class ${className}
 * 
 *  silecsData.${className}_${deviceLabel}.${blockName}.${registerName} = ....;
 */""")

blockModeDeviceListExample = string.Template("""
 * This example shows how to address the register ${registerName} of block ${blockName}
 * of device ${deviceLabel} of the class ${className}
 *
 *  silecsData.${className}_${blockName}.${deviceLabel}.${registerName} = ....;
 */""")

#=========================================================================
# C code generation Sub-function
#=========================================================================

# HEADER file ------------------------------------------------------------
def cHeader(silecsRelease, baseAddress):
    return header.substitute(silecsRelease=silecsRelease, baseAddress=baseAddress)

# REGISTER definition ----------------------------------------------------
def cRegister(regName, regFormat, regDim, regDim2=1, strLen=1):
    if strLen > 1: # if it's a string
        if regDim2 > 1:    # add another pair of brackets to hold the string length in case second dimension of array is set
            return stringArray.substitute(regFormat=whichDIGIFormat[regFormat], regName=regName, regDim=regDim, regDim2=regDim2, strLen=strLen)
        elif regDim > 1:    # reuse the array2D syntax
            return regArray2d.substitute(regFormat=whichDIGIFormat[regFormat], regName=regName, regDim=regDim, strLen=strLen)
        else:        # regular string register
            return regArray.substitute(regFormat=whichDIGIFormat[regFormat], regName=regName, strLen=strLen)
    else:
        if regDim == 1 and regDim2 == 1:    # scalar
            return regScalar.substitute(regFormat=whichDIGIFormat[regFormat], regName=regName)
        elif regDim > 1 and regDim2 == 1:    # array
            return regArray.substitute(regFormat=whichDIGIFormat[regFormat], regName=regName, strLen=regDim)
        else:    # dim1>=1 and for whatever dim2, use double array syntax
            return regArray2d.substitute(regFormat=whichDIGIFormat[regFormat], regName=regName, regDim=regDim, strLen=regDim2)


def cFirstBlockUDT(className, classVersion):
    return firstBlockUDT.substitute(className=className, classVersion=classVersion)
    
    
# BLOCK type definition (UDT) --------------------------------------------
def cBlockUDT(className, blockName, regList):
    return blockUDT.substitute(regList=regList, className=className, blockName=blockName)
    

#============= DEVICE instantiation ============= 
   
# GLOBAL instantiation ----------------------------------------------

def cAllocationComment(plcProtocol):
    return allocationComment.substitute(plcProtocol=plcProtocol)

     
# DEVICE MODE instantiation ----------------------------------------------

def cDeviceModeBlockInstantiation(className, blockName):
    return deviceModeBlockInstantiation.substitute(className=className, blockName=blockName)

def cDeviceModeClass_deviceNumber(className, blockList):
    return deviceModeClass_deviceNumber.substitute(blockList=blockList, className=className)

def cDeviceModeClass_deviceList(blockList, deviceList):
    return deviceModeClass_deviceList.substitute(blockList=blockList, deviceList=deviceList)

def cDeviceModeDataInstantiation(classList):
    return deviceMode_dataInstantiation.substitute(classList=classList)

# BLOCK MODE instantiation ----------------------------------------------

def cBlockModeDeviceInstantiation_deviceList(className, blockName, deviceName):
    return blockModeDeviceInstantiation_deviceList.substitute(className=className, blockName=blockName, deviceName=deviceName)

def cBlockModeBlockInstantiation(deviceList, className, blockName):
    return blockModeBlockInstantiation.substitute(deviceList=deviceList, className=className, blockName=blockName)
    
def cBlockModeDataInstantiation(classList):
    return blockMode_dataInstantiation.substitute(classList=classList)

# Init Function ----------------------------------------------
def cInitDeviceMode(ieVersion,ieChecks,ieUser):
    dt = time.localtime(time.time())
    return initFunctionDeviceMode.substitute(version="SILECS_"+ieVersion, checksum=ieChecks, user=ieUser, dt6=dt[6], dt5=dt[5], dt4=dt[4], dt3=dt[3], dt2=dt[2], dt1=dt[1], dt0=dt[0])

def cInitBlockMode(ieVersion,ieChecks,ieUser):
    dt = time.localtime(time.time())
    return initFunctionBlockMode.substitute(version="SILECS_"+ieVersion, checksum=ieChecks, user=ieUser, dt6=dt[6], dt5=dt[5], dt4=dt[4], dt3=dt[3], dt2=dt[2], dt1=dt[1], dt0=dt[0])
        
# Example function  ----------------------------------------------
def cExample(plcProtocol,className, blockName, registerName, deviceLabel =''):
    if plcProtocol == 'DEVICE_MODE':        
        return instantiationExample  + deviceModeDeviceListExample.substitute(registerName=registerName, blockName=blockName, deviceLabel=deviceLabel, className=className)
    else:
        return instantiationExample + blockModeDeviceListExample.substitute(registerName=registerName, blockName=blockName, deviceLabel=deviceLabel, className=className)
